#include "polygraph.h"

/*
 * $Id: SrvConnMgr.cc,v 1.4 1999/05/26 19:03:20 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "Connection.h"
#include "LogComment.h"
#include "SharedOpts.h"
#include "SrvOpts.h"
#include "SrvConnMgr.h"
#include "polyErrors.h"



SrvConnMgr::SrvConnMgr(User *aUser): theUser(aUser) {
	Assert(theUser);
	theConns.limit(Socket::TheMaxLevel);
	theConnIdx.stretch(Socket::TheMaxLevel);

	if (TheSrvOpts.trackRPorts) {
		theRPortStamps.stretch(65*1024);
		theRPortStamps.count(theRPortStamps.capacity());
		theRPortStamps.memset(0);
	}
}

SrvConnMgr::~SrvConnMgr() {
}

void SrvConnMgr::put(Connection *conn) {
	Assert(conn);
	Assert(theConnIdx[conn->sock().fd()] == conn);

	// see if we want to keep this connection open
	if (conn->reusable())
		keepOpen(conn);
	else
		close(conn);
}

Connection *SrvConnMgr::get(Socket &s, bool &fatal) {

	Connection *conn = theConns.get();
	if (!conn->accept(s, fatal)) {
		theConns.put(conn); // no need to close a non-opened conn
		return 0;
	}

	const int fd = conn->sock().fd();
	Assert(!theConnIdx[fd]);
	theConnIdx[fd] = conn;
	setUseLimit(conn);
	conn->use();
	justOpened(conn);
	//cerr << here << "returned fd: " << fd << " conn: " << (void*)conn << endl;
	return conn;
}

// internal method to close a connection, no checks made here
void SrvConnMgr::close(Connection *conn) {
	//cerr << here << "closing fd: " << conn->sock().fd() << " conn: " << (void*)conn << endl;
	aboutToClose(conn);
	theConnIdx[conn->sock().fd()] = 0;
	conn->close();
	theConns.put(conn);
}

void SrvConnMgr::noteReadReady(int fd) {
	Connection *conn = theConnIdx[fd];
	Assert(conn && conn->sock().fd() == fd);
	conn->theRd.stop(this);
	conn->use();
	theUser->noteConnReady(conn);
}

void SrvConnMgr::keepOpen(Connection *conn) {
	Assert(!conn->theRd.theReserv && !conn->theWr.theReserv);
	if (conn->theRdBuf.contSize()) {
		conn->use();
		theUser->noteConnReady(conn);
	} else
		conn->theRd.start(this);
}

void SrvConnMgr::setUseLimit(Connection *conn) {
	int limit = 1; // no pconns by default
	if (RndDistr *d = TheOpts.thePconnUseLmt.condDistr()) {
		// zero limit means default case
		limit = (int)d->trial();
		if (limit <= 0)
			limit = 1;
	}
	conn->useLimit(limit);
}

void SrvConnMgr::justOpened(Connection *conn) {
	Assert(conn);

	if (theRPortStamps.count()) {
		const int rport = conn->rport();

		if (rport < 0 || rport > theRPortStamps.count()) {
			ReportError(rport < 0 ? errBadRPort : errRPortOutOfRange);
			return;
		}

		const int lastAccessSecs = theRPortStamps[rport];
		
		// check that the port is inactive
		if (lastAccessSecs > 0) {
			ReportError(errRPortUsedTwice);
			// reset or we  would think there is an internal 
			// accounting error!
			theRPortStamps[rport] = -lastAccessSecs;
			return;
		}

		static int ConnCount = 0; // XXX: should maintain a global counter?
		ConnCount++;

		// if first access do nothing
		if (lastAccessSecs == 0)
			return;

		theRPortStamps[rport] = -lastAccessSecs; // mark "in use";

		// measure the idle time
		const Time lastAccess = TheClock.TheStartTime + Time::Sec(-lastAccessSecs);
		const int waitSecs = (TheClock - lastAccess).sec();

		// report first re-use
		static int ReuseCnt = 0;
		if (!ReuseCnt++)
			Comment(2) << "first rport reuse: port " << rport << " within " << waitSecs << " sec; after " << ConnCount << " conns" << endc;

		// warn if a new low record
		static int MinWaitSecs = TheSrvOpts.theMinRPortWait.sec();
		if (waitSecs < MinWaitSecs) {
			Comment(2) << "rport " << rport << " reused within " << waitSecs << " sec (old record: " << MinWaitSecs << " ); after " << ConnCount << " conns" << endc;
			MinWaitSecs = waitSecs;
		}
	}
}

void SrvConnMgr::aboutToClose(Connection *conn) {
	Assert(conn);
	if (theRPortStamps.count()) {
		const int rport = conn->rport();
		if (rport < 0 || rport > theRPortStamps.count())
			return; // we probably have complained already

		// check that the port is active
		Assert(theRPortStamps[rport] >= 0);

		// record the time of close
		theRPortStamps[rport] = -(TheClock - TheClock.TheStartTime).sec();
	}
}
